/*
 * Copyright(c) 2010, Norio Kobota, All rights reserved.
 */

#include "mod_websocket.h"
#include "mod_websocket_socket.h"

# include <string.h>

#ifdef HAVE_PCRE_H
# include <pcre.h>
#endif /* HAVE_PCRE_H */

# include <limits.h>

# ifdef HAVE_STDINT_H
#  include <stdint.h>
# endif

# ifdef USE_OPENSSL
#  include <openssl/sha.h>
typedef unsigned char sha1_byte;
# else
#  include "mod_websocket_sha1.h"
#  define	SHA_DIGEST_LENGTH	SHA1_DIGEST_LENGTH
# endif

# include "mod_websocket_base64.h"

static mod_websocket_bool_t is_allowed_origin(handler_ctx *hctx) {
    size_t i;
    data_unset *du = NULL;
    array *allowed_origins = NULL;
    data_string *allowed_origin = NULL;

#ifdef	HAVE_PCRE_H
    pcre *re = NULL;
    int rc = 0;
    const char *err_str;
    int err_off;
# define	N	(10)
    int ovec[N * 3];
#endif	/* HAVE_PCRE_H */

    du = array_get_element(hctx->ext->value, "origins");
    if (!du || du->type != TYPE_ARRAY) {
        DEBUG_LOG(MOD_WEBSOCKET_LOG_INFO, "s", "allowed origins are not specified");
        return MOD_WEBSOCKET_TRUE;
    }
    allowed_origins = ((data_array *)du)->value;
    if (!allowed_origins || allowed_origins->used == 0) {
        DEBUG_LOG(MOD_WEBSOCKET_LOG_INFO, "s", "allowed origins are not specified");
        return MOD_WEBSOCKET_TRUE;
    }
    if (buffer_is_empty(hctx->handshake.origin)) {
        DEBUG_LOG(MOD_WEBSOCKET_LOG_INFO, "s", "Origin header does not exist in client request, but origins are specified");
        return MOD_WEBSOCKET_FALSE;
    }
    for (i = allowed_origins->used; i > 0; i--) {
        du = allowed_origins->data[i - 1];
        if (du->type != TYPE_STRING) {
            DEBUG_LOG(MOD_WEBSOCKET_LOG_WARN, "s", "allowed origin configuration is invalid");
            continue;
        }
        allowed_origin = (data_string *)du;
        if (buffer_is_empty(allowed_origin->value)) {
            DEBUG_LOG(MOD_WEBSOCKET_LOG_WARN, "s", "allowed origin value is empty");
            continue;
        }

#ifdef	HAVE_PCRE_H
        re = pcre_compile(allowed_origin->value->ptr, 0, &err_str, &err_off, NULL);
        if (!re) {
            DEBUG_LOG(MOD_WEBSOCKET_LOG_WARN, "ss", allowed_origin->value->ptr, "is invalid RegExp");
            continue;
        }
        rc = pcre_exec(re, NULL, hctx->handshake.origin->ptr, hctx->handshake.origin->used - 1,
                       0, PCRE_ANCHORED, ovec, N);
        free(re);
        if (rc > 0) {
            DEBUG_LOG(MOD_WEBSOCKET_LOG_INFO, "sss",
                      hctx->handshake.origin->ptr, "is match allowed origin:", allowed_origin->value->ptr);
            return MOD_WEBSOCKET_TRUE;
        }
# undef	N
#else
        if (buffer_is_equal(allowed_origin->value, hctx->handshake.origin)) {
            DEBUG_LOG(MOD_WEBSOCKET_LOG_INFO, "sss",
                      hctx->handshake.origin->ptr, "is match allowed origin:", allowed_origin->value->ptr);
            return MOD_WEBSOCKET_TRUE;
        }
#endif	/* HAVE_PCRE_H */

    }
    DEBUG_LOG(MOD_WEBSOCKET_LOG_INFO, "ss",
              hctx->handshake.origin->ptr, "is not match any allowed origins");
    return MOD_WEBSOCKET_FALSE;
}

mod_websocket_errno_t mod_websocket_handshake_check_request(handler_ctx *hctx) {
    size_t i;
    array *hdrs;
    data_string *hdr = NULL;
    mod_websocket_handshake_t *handshake;
    buffer *version_hdr_value = NULL;

    if (hctx == NULL || hctx->con == NULL) {
        return MOD_WEBSOCKET_INTERNAL_SERVER_ERROR;
    }
    hdrs = hctx->con->request.headers;
    handshake = &hctx->handshake;

    /* store necessary headers */
    for (i = hdrs->used; i > 0; i--) {
        hdr = (data_string *)hdrs->data[i - 1];
        if (buffer_is_equal_string(hdr->key, CONST_STR_LEN("Host"))) {
            handshake->host = hdr->value;
        }
        if (buffer_is_equal_string(hdr->key, CONST_STR_LEN("Sec-WebSocket-Version"))) {
            version_hdr_value = hdr->value;
        }
        if (buffer_is_equal_string(hdr->key, CONST_STR_LEN("Origin")) ||
            buffer_is_equal_string(hdr->key, CONST_STR_LEN("Sec-WebSocket-Origin"))) {
            handshake->origin = hdr->value;
        }

        if (buffer_is_equal_string(hdr->key, CONST_STR_LEN("Sec-WebSocket-Key"))) {
            handshake->key = hdr->value;
        }

    }
    if (buffer_is_empty(handshake->host)) {
        DEBUG_LOG(MOD_WEBSOCKET_LOG_ERR, "s", "Host header does not exist");
        return MOD_WEBSOCKET_BAD_REQUEST;
    }
    /*if (buffer_is_empty(handshake->origin)) {
        DEBUG_LOG(MOD_WEBSOCKET_LOG_ERR, "s", "Origin header does not exist");
        return MOD_WEBSOCKET_BAD_REQUEST;
    }*/
    if (is_allowed_origin(hctx) != MOD_WEBSOCKET_TRUE) {
        return MOD_WEBSOCKET_FORBIDDEN;
    }
    if (buffer_is_empty(version_hdr_value)) {
        handshake->version = 0;
    } else {
        handshake->version = (int)(strtol(version_hdr_value->ptr, NULL, 10) & INT_MAX);
    }

    return MOD_WEBSOCKET_OK;
}

static mod_websocket_errno_t create_response_rfc_6455(handler_ctx *hctx) {
    const char *const_hdr = "HTTP/1.1 101 Switching Protocols\r\n"
                            "Upgrade: websocket\r\n"
                            "Connection: Upgrade\r\n";
    buffer *key;
    buffer *resp = NULL;
    SHA_CTX sha;
    unsigned char sha_digest[SHA_DIGEST_LENGTH];
    unsigned char *accept_body;
    size_t accept_body_siz;
    data_unset *du;

    if (buffer_is_empty(hctx->handshake.key)) {
        DEBUG_LOG(MOD_WEBSOCKET_LOG_ERR, "s", "Sec-WebSocket-Key is invalid");
        return MOD_WEBSOCKET_BAD_REQUEST;
    }

    key = buffer_init_buffer(hctx->handshake.key);
/* refer: RFC-6455 Sec.1.3 Opening Handshake */
#define	GUID	"258EAFA5-E914-47DA-95CA-C5AB0DC85B11"
    buffer_append_string(key, GUID);
#undef GUID

    /* get SHA1 hash of key */
    SHA1_Init(&sha);
    SHA1_Update(&sha, (sha1_byte *)key->ptr, key->used - 1);
    SHA1_Final(sha_digest, &sha);
    buffer_free(key);

    /* get base64 encoded SHA1 hash */
    if (mod_websocket_base64_encode(&accept_body, &accept_body_siz, sha_digest, SHA_DIGEST_LENGTH) < 0) {
        DEBUG_LOG(MOD_WEBSOCKET_LOG_ERR, "s", "no memory");
        return MOD_WEBSOCKET_INTERNAL_SERVER_ERROR;
    }

    resp = chunkqueue_get_append_buffer(hctx->tocli);
    if (resp == NULL) {
        free(accept_body);
        DEBUG_LOG(MOD_WEBSOCKET_LOG_ERR, "s", "no memory");
        return MOD_WEBSOCKET_INTERNAL_SERVER_ERROR;
    }
    buffer_append_string(resp, const_hdr);
    buffer_append_string(resp, "Sec-WebSocket-Accept: ");
    buffer_append_string_len(resp, (char *)accept_body, accept_body_siz);
    free(accept_body);
    buffer_append_string(resp, "\r\n");

    du = array_get_element(hctx->ext->value, "subproto");
    if (du && du->type == TYPE_STRING) {
        buffer_append_string(resp, "Sec-WebSocket-Protocol: ");
        buffer_append_string_buffer(resp, ((data_string *)du)->value);
        buffer_append_string(resp, "\r\n");
    }
    buffer_append_string(resp, "\r\n");
    hctx->mode = MOD_WEBSOCKET_WEBSOCKET_PROXY;
    return MOD_WEBSOCKET_OK;
}

mod_websocket_errno_t mod_websocket_handshake_create_response(handler_ctx *hctx) {
    if (!hctx) {
        return MOD_WEBSOCKET_INTERNAL_SERVER_ERROR;
    }
    DEBUG_LOG(MOD_WEBSOCKET_LOG_DEBUG, "s", "send handshake response");

    if (hctx->handshake.version >= 8) {
        return create_response_rfc_6455(hctx);
    }
    DEBUG_LOG(MOD_WEBSOCKET_LOG_ERR, "s", "not supported WebSocket Version");
    return MOD_WEBSOCKET_SERVICE_UNAVAILABLE;
}

